from cosmosis.runtime import FunctionModule
from cosmosis.datablock import DataBlock
from cosmosis.gaussian_likelihood import GaussianLikelihood
import numpy as np
import os

def test_gaussian():
    class MyLikelihood(GaussianLikelihood):
        x_section = "aaa"
        x_name = "a"
        y_section = "bbb"
        y_name = "b"
        like_name = "lll"

        def build_data(self):
            x_obs = np.array([1.0, 2.0, 3.0])
            y_obs = x_obs * 2
            return x_obs, y_obs

        def build_covariance(self):
            covmat = np.diag([0.1, 0.1, 0.1])
            return covmat

    mod = MyLikelihood.as_module("my")

    # no extra config info
    mod.setup({"my":{"include_norm":True}})

    block = DataBlock()
    block["aaa", "a"] = np.arange(5.)
    block["bbb", "b"] = np.arange(5.) * 2
    status = mod.execute(block)

    assert status == 0
    assert np.isclose(block["data_vector", "lll_chi2"], 0)
    assert np.isclose(block["data_vector", "lll_log_det"], 3*np.log(0.1))
    assert block["data_vector", "lll_n"] == 3

    assert np.isclose(block["likelihoods", "lll_like"], -3*np.log(0.1)/2)



if __name__ == '__main__':
    test_gaussian()





import os
import yaml
import glob
from collections import OrderedDict

#Generate the summary page of all the samplers
#Generate the specific page for each sampler.

def name_for_sampler_page(name):
	return name

page_template = u"""
# The {name} sampler

## {purpose}

Name: **{name}**

Version: **{version}**

Author(s): **{authors}**

URL: **{url}**

Cite: ** {citations} **

Parallel: **{parallel}**


{explanation}

##Installation

{installation}


##Parameters

These parameters can be set in the sampler's section in the ini parameter file.  
If no default is specified then the parameter is required. A listing of "(empty)" means a blank string is the default.

Parameter | Type | Meaning | Default
------------|------|---------|--------
{parameter_lines}

"""

def parse_parameter_description(desc):
	paren, rest = desc.split(')', 1)
	paren = paren.strip()
	rest = rest.strip().strip('"')
	paren=paren.lstrip('(')
	if ';' in paren:
		dtype, default=paren.split(';')
		default=default.split('=')[1]
	else:
		dtype=paren
		default=""
	if default.strip()=="''":
		default = '(empty)'
	return dtype, default, rest



def generate_sampler_wiki(info):
	"Generate wiki markdown for a single sampler"
	info = info.copy()
	name = info['name']
	info['explanation'] = info['explanation'].replace("\n",'\n\n').strip('"')
	info['installation'] = info['installation'].replace("\n",'\n\n').strip('"')
	page_name = name_for_sampler_page(name)
	page = open('wiki/{}.md'.format(page_name), 'w')
	parameter_lines = []
	for pname,description in list(info['params'].items()):
		try:
			dtype, default, rest = parse_parameter_description(description)
			parameter_lines.append("{}|{}|{}|{}".format(pname, dtype, rest, default))
		except (IndexError, ValueError):
			print("ERROR: Could not parse in {0}".format(name))
			print(description)
			continue
		
	parameter_lines = '\n'.join(parameter_lines)
	info['name'] = name.capitalize()
	markdown = page_template.format( 
		citations=', '.join(info['cite']),
		authors = ','.join(info['attribution']),
		parameter_lines=parameter_lines,
		**info
		)
	page.write(markdown)
	page.close()



def generate_overview(infos):
	pass


def generate_links(infos):
	for info in infos:
		name = info['name']
		page = name_for_sampler_page(name)
		slug = info['purpose']
		print(" - [{0} sampler](samplers/{1}) {2}".format(page, name, slug))


def main():
	#get the base dir to work from
	src=os.environ['COSMOSIS_SRC_DIR']
	sampler_dir=os.path.join(src, "cosmosis", "samplers")
	#Find and parse all the files
	search_path = "{}/*/sampler.yaml".format(sampler_dir)
	yaml_files = glob.glob(search_path)
	infos = [yaml.safe_load(open(f)) for f in yaml_files]
	#Make the ordering the same every time
	try:
		os.mkdir('wiki')
	except OSError:
		pass
	generate_links(infos)
	for info in infos:
		generate_sampler_wiki(info)


if __name__ == '__main__':
	main()


from cosmosis.postprocessing.postprocess import postprocessor_for_sampler
from cosmosis.runtime.config import Inifile
from cosmosis.runtime.parameter import Parameter
import sys

# Set up a quick test example
pp_class = postprocessor_for_sampler('emcee')
ini1=Inifile("./demos/demo5.ini")
postprocessor = pp_class(ini1, "demo5", 0)


# actual example
values_ini = postprocessor.extract_ini("VALUES")
params = Parameter.load_parameters(values_ini)

for param in params:
    print(param, param.limits)



from cosmosis import DataBlock
from cosmosis.datablock.cosmosis_py.errors import BlockSectionNotFound
import numpy as np

def setup(options):
	return 1

def execute(block, config):
	x = np.array([-1.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.,-10.])
	y = np.array([-2.,-4.,-8.,-16.,-32.])
	nx = len(x)
	ny = len(y)
	z = np.zeros([nx,ny])
	for i in range(nx):
		for j in range(ny):
			z[i,j] = 10*i+j
	block.put_grid("py_put", "x", x, "y", y, "z", z)

	try:
		x_c, y_c, z_c = block.get_grid("c_put", "x", "y", "z")
		assert (x==x_c).all()
		assert (y==y_c).all()
		assert (z==z_c).all()
	except BlockSectionNotFound:
		print("No C in py")

	try:
		x_f90, y_f90, z_f90 = block.get_grid("f90_put", "x", "y", "z")
		assert (x==x_f90).all()
		assert (y==y_f90).all()
		assert (z==z_f90).all()
	except BlockSectionNotFound as err:
		print("No f90 in py ", err)
	return 0

def cleanup(config):
	pass


"""

You don't need to use this script.

It is for the developers to make header files.

"""

print(__doc__)
import sys
#You don't need to use this file - it is for the developers to 
try:
	import astropy.constants
	import tabulate
except ImportError:
	print("I can tell you aren't really a developer because you don't have")
	print("the dependencies (tabulate, astropy).")

def find_constants():
	consts = []
	for x in dir(astropy.constants):
		c = getattr(astropy.constants, x)
		if isinstance(c, astropy.constants.constant.Constant):
			consts.append(c)
	return consts

class ConstantFile(object):
	def write(self, constants):
		lines = []
		for c in constants:
			parts = [
				"{self.prefix}{c.abbrev}".format(self=self,c=c),
				"=",
				"{}".format(float(c.value)),
				"{self.comment} {c.unit}".format(self=self,c=c),
				"{self.comment} {c.name}".format(self=self,c=c),
			]
			lines.append(parts)
		f = open(self.filename, 'w')
		f.write(self.header)
		f.write(tabulate.tabulate(lines,tablefmt="plain", numalign='left',floatfmt="e"))
		f.write("\n")
		f.write(self.footer)
		f.write("\n")
		f.close()



class Python(ConstantFile):
	filename = "cosmosis/constants.py"
	header = ""
	prefix = ""
	comment = "#"
	footer = ""

class C(ConstantFile):
	filename = "cosmosis/datablock/cosmosis_constants.h"
	header = "#ifndef COSMOSIS_CONSTANTS_H\n#define COSMOSIS_CONSTANTS_H\n"
	prefix = "static const double constant_"
	comment = "//"
	footer = "#endif /* COSMOSIS_CONSTANTS_H */"

class Fortran(ConstantFile):
	filename = "cosmosis/datablock/cosmosis_constants.fh"
	header = ""
	prefix = "real(8), parameter ::  constant_"
	comment = "!"
	footer = ""




def main():
	constants = find_constants()
	for cls in [Python, C, Fortran]:
		cls().write(constants)


if __name__ == '__main__':
	main()


"""
This script, when run from the cosmosis-standard-library directory, 
scans for module.yaml files and uses them to generate the markdown 
for the wiki page documenting the standard modules.

It is not intended for users.

"""
import yaml
import os
import collections
import sys
import codecs
# UTF8Writer = codecs.getwriter('utf8')
# sys.stdout = UTF8Writer(sys.stdout)

#From stack overflow!
def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=collections.OrderedDict):
	class OrderedLoader(Loader):
		pass
	OrderedLoader.add_constructor(
		yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
		lambda loader, node: object_pairs_hook(loader.construct_pairs(node)))
	return yaml.load(stream, OrderedLoader)

page_template = u"""
# {name} module

## {purpose}

**Name**: {name}

**File**: {filepath}

**Version**: {version}

**Author(s)**:

{author_list}

**URL**: {url}

**Cite**: 

{cite_list}

**Rules**: 

{rule_list}

**Assumptions**:

{assumption_list}

**Explanation**

{explanation}


##Parameters

These parameters can be set in the module's section in the ini parameter file.  
If no default is specified then the parameter is required.

Parameter | Description
------------|-------
{parameter_lines}

##Inputs

These parameters and data are inputs to the module, either supplied as parameters by the sampler or
computed by some previous module.  They are loaded from the data block.

Section | Parameter | Description
--------|-----------|-------
{input_lines}

##Outputs

These parameters and data are computed as outputs from the module

Section | Parameter | Description
--------|-----------|-------
{output_lines}


"""

#find the modules
outputs = collections.OrderedDict()
for dirpath, dirnames, filenames in os.walk('.'):
	if dirpath=='.': continue
	for filename in filenames:
		if filename=='module.yaml':
			category, module_name = os.path.split(dirpath)
			category = os.path.split(category)[1]
			if category=='.':continue
			sys.stderr.write("Found %s/%s\n"%(category,module_name))
			filepath = os.path.join(dirpath, filename)
			if category not in outputs:
				outputs[category] = collections.OrderedDict()
			info = ordered_load(open(filepath))
			name = info['name']
			version = info['version']
			info['full_name'] = '{}--{}'.format(name,version)
			info['page_name'] = '{}_{}'.format(name,version)
			if dirpath.startswith('./'): dirpath=dirpath[2:]
			info['dirname'] = dirpath
			info['explanation'] = info['explanation'].strip().strip('"').replace("\n", "  \n")#.replace("\n","\n\n")
			outputs[category][info['full_name']] = info


f = open("wiki/default_modules.md","w")
for cat in outputs:
	f.write(u"## {}\n\n".format(cat))
	for mod in outputs[cat]:
		page_name = outputs[cat][mod]['page_name']
		purpose = outputs[cat][mod]['purpose']
		version = outputs[cat][mod]['version']
		name = outputs[cat][mod]['name']
		f.write(u"- [{} ({})](https://bitbucket.org/joezuntz/cosmosis/wiki/default_modules/{})  {}\n\n".format(name,version,page_name,purpose))
f.close()


def make_list(l):
	if l is None: return ""
	if isinstance(l, str):
		return u"- "+l
	else:
		return u"\n".join("- "+m for m in l)

def make_params(P):
	lines = []
	for section, params in list(P.items()):
		for i,(name,desc) in enumerate(params.items()):
			sec = section if i==0 else ""
			lines.append(u"{}|{}|{}".format(sec,name,desc))
	return "\n".join(lines)

def make_page(info):
	parameter_lines = []
	for name,desc in list(info['params'].items()):
		parameter_lines.append(u"{}|{}".format(name,desc))
	parameter_lines = '\n'.join(parameter_lines)
	input_lines = make_params(info['inputs'])
	output_lines = make_params(info['outputs'])
	markdown = page_template.format(
		author_list = make_list(info['attribution']),
		cite_list = make_list(info['cite']),
		rule_list = make_list(info['rules']),
		assumption_list = make_list(info['assumptions']),
		input_lines = input_lines,
		output_lines = output_lines,
		parameter_lines=parameter_lines,
		filepath=info['dirname'] + "/" + info['interface'],
		**info)
	open("wiki/default_modules/{}.md".format(info['page_name']),"wb").write(markdown.encode("utf-8"))



for cat in outputs:
	for mod in outputs[cat]:
		try:
			make_page(outputs[cat][mod])
		except:
			print(cat, mod)
			raise



from numpy import exp

def setup(options):
	return []


def execute(block, config):
	x = block['banana', 'x']
	y = block['banana', 'y']
	z = y**2/0.6**2 + (x-y**2)**2/0.1**2
	#like = exp(-0.5*z) * exp(-0.5*x**2) * exp(-0.5*y**2) 
	like = - 0.5*z
	block['likelihoods', 'banana_like'] = like
	return 0


from cosmosis.samplers import ParallelSampler
import numpy as np
import sys


def task(p):
    r = pipeline.run_results(p)
    return r.vector, r.extra, r.prior, r.post

# This (not very useful, except possibly for debugging)
# sampler draws a line between the two
# extreme corners of the distribution and evaluates points
# along it.

class CornerLineSampler(ParallelSampler):
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        self.ndim = self.pipeline.nvaried
        self.nsample = self.read_ini('nsample', int)
        corner1 = np.array([p.limits[0] for p in self.pipeline.varied_params])
        corner2 = np.array([p.limits[1] for p in self.pipeline.varied_params])
        x = np.linspace(0, 1, self.nsample)
        self.samples = corner1 + x[:, np.newaxis] * (corner2 - corner1)
        self.complete = False


    def execute(self):
        # Evaluate points optionally in parallel
        jobs = list(self.samples)

        if self.pool:
            results = self.pool.map(task, jobs)
        else:
            results = list(map(task, jobs))

        for (vector, extra, prior, post) in results:
            self.output.parameters(vector, extra, prior, post)
        self.complete = True

    def is_converged(self):
        return self.complete
